/*
 * Copyright (c) 1995, Jack F. Vogel
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Jack F. Vogel
 * 4. The name of the developer may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * mpcore.s:	FreeBSD machine support for the Intel MP Spec
 *		multiprocessor systems.
 */

#include "opt_smp.h"

#include "assym.s"			/* system definitions */
#include <machine/specialreg.h>		/* x86 special registers */
#include <machine/asmacros.h>		/* miscellaneous asm macros */
#include <machine/smpasm.h>
#include <machine/mpasm.h>

/*
 * Globals
 */
	.globl _mpenabled, _mpfps
_mpenabled:	.long  0		
_mpfps: 	.long  0		

	.text
/*
 * Multiprocessor Initialization
 */

NON_GPROF_ENTRY(apic_probe)
/*
 * The Intel MP Spec calls for the comforming machine to
 * have a structure in memory to be searched for in the
 * following order:
 *	1) If the machine has EBDA, then in the FIRST
 *	   1K of that.
 *	2) ELSE (that is with NO EBDA) in the LAST
 *	   1K of REAL-MODE memory (machine dependent).
 *	3) Or finally in the BIOS itself.
 */
	/* Do we have EBDA? */
	movw	0x40e, %ax 	/* 40:0e in real-mode terms */
	cmpw	$0, %ax		/* If implemented, its non-zero */
	jz	mem		/* NO, go search memory */
	shll	$4, %eax	/* Convert to physical address */
	movl	%eax, %esi
	movl	%esi, %edi
	addl	$0x400, %edi	/* Add 1K - end point */
	movl	$MP_SIG, %eax	/* This is the signature */
1:	movl	(%esi), %ebx
	cmpl	%ebx, %eax
	jz	match		/* We got a match */
	inc	%esi
	cmpl	%esi, %edi
	jnz	1b
	jmp	bios		/* It may still be there */

mem:
/* 
 * If I wanted to be COMPLETELY thorough I should interrogate
 * to find out if the machine REALLY had 640K base, but at this
 * point that is just ANAL. By fiat I declare it MUST have 640
 * to be supported. :-)
 */
	movl	$0x9fc00, %esi	/* base of the last 1K of 640K */
	movl	$0x9ffff, %edi
	movl	$MP_SIG, %eax	/* This is the signature */
1:	movl	(%esi), %ebx
	cmpl	%ebx, %eax
	jz	match		/* We got a match */
	inc	%esi
	cmpl	%esi, %edi
	jnz	1b
	/* Fall through if this fails... */

bios:	/* Search the Bios */
	movl	$MP_SIG, %eax
	movl	$0xf0000, %esi
	movl	$0xfffff, %edi
1:	movl	(%esi), %ebx
	cmpl	%ebx, %eax
	jz	match		/* We got a match */
	inc	%esi
	cmpl	%esi, %edi
	jnz	1b
	ret		/* Not found, not MP capable */
match:	movl	$1, _mpenabled-KERNBASE
	movl	%esi, _mpfps-KERNBASE	/* Store the physical address */

/* TODO: We need code that will analyse the MP structures
 * and do the fancy potential configuration. For now I am
 * assuming a simple DUAL processor.
 */

/* XXX: remember, before paging */

	/* Enable the APIC, and configure interrupts */
	movl	$(APIC_BASE+APIC_SVR), %esi	/* Spurious vector register */
	movl	(%esi), %eax
	orl	$0x00000100, %eax		/* This bit enables the APIC */
	movl	%eax,(%esi)

	movl	$(APIC_BASE+APIC_LVT1), %esi	/* Setup LVT1 as ExtINT */
	movl	(%esi), %eax
	andl	$0xfffe58ff, %eax
	orl	$0x700, %eax
	movl	%eax, (%esi)

	movl	$(APIC_BASE+APIC_LVT2), %esi	/* Setup LVT2 as NMI */
	movl	(%esi), %eax
	andl	$0xfffe58ff, %eax
	orl	$0xffff0400, %eax
	movl	%eax, (%esi)
	ret
	
NON_GPROF_ENTRY(apic_installtramp)
/*
 * Due to the way Intel devised MP processor startup we
 * need a real mode 'trampoline' to get the other processors
 * started, in fact we can't even be sure the other processors
 * are out there without doing it.  What this small chunk
 * of code will do is put the processor in protected mode
 * and then have it do an intersegment jump into kernel 
 * physical memory above 1Meg.
 *
 * KEEP IN MIND, at this point on the booting processor
 * we still have not enabled paging.
 */
	movl	_bootMP_size-KERNBASE, %ecx
	lea	_bootMP-KERNBASE,%esi		/* source */
	movl	$MP_INITADDR,%edi		/* destination */
	cld
	rep
	movsb
	ret

/*
#define CPUNBR 0x00000000
*/

#ifndef CPUNBR 
#define CPUNBR 0x01000000
#endif
NON_GPROF_ENTRY(apic_startup)
/* 
 * This routine goes through a sequence of InterProcessor 
 * Interrupts that will get the secondary processor(s)
 * started. I realize its cryptic, for elaboration see
 * the MP Spec.		-jackv
 */
	movl	_mpenabled, %eax /* See if the probe succeeded */
	cmpl	$1, %eax
	jz	1f
	ret		/* SMP not enabled */
1:
	pushl	%esi
	pushl	%ebx

	call	_get_mplock

	movl	_apic_base, %esi

/* Step 1 - Do a INIT/Reset seqeuence */

	movl	APIC_ICR_HI(%esi),%eax
	andl	$0xf0ffffff,%eax /* mask out ID bits */
	orl	$CPUNBR,%eax
	movl	%eax,APIC_ICR_HI(%esi)
	movl	APIC_ICR_LOW(%esi),%eax
	andl	$0xfff00000,%eax
	orl	$0x0000c500,%eax	/* do a INIT IPI - ASSERT/RESET */
	movl	%eax,APIC_ICR_LOW(%esi)

3:	movl	APIC_ICR_LOW(%esi),%eax	/* wait for pending status end */
	andl 	$0x00001000,%eax
	jnz 	3b

	movl	APIC_ICR_HI(%esi),%eax
	andl	$0xf0ffffff,%eax /* mask out ID bits */
	orl	$CPUNBR,%eax
	movl	%eax,APIC_ICR_HI(%esi)
	movl	APIC_ICR_LOW(%esi),%eax
	andl	$0xfff00000,%eax
	orl 	$0x00008500,%eax	/* do a INIT IPI - DEASSERT/RESET */
	movl	%eax,APIC_ICR_LOW(%esi)

3:	movl	APIC_ICR_LOW(%esi),%eax	/* wait for pending status end */
	andl 	$0x00001000,%eax
	jnz 	3b

/* Step 2 - Do the STARTUP IPI sequence */

	xorl	%ebx, %ebx
	movl	$0x1000, %eax /* Delay a bit */
1:	decl	%eax
	cmpl	%eax, %ebx
	jnz	1b

2:	movl	APIC_ICR_LOW(%esi),%eax	/* wait for pending status end */
	andl 	$0x00001000,%eax
	jnz 	2b

	movl	APIC_ICR_HI(%esi),%eax
	andl	$0xf0ffffff,%eax /* mask out ID bits */
	orl	$CPUNBR,%eax
	movl	%eax,APIC_ICR_HI(%esi)
	movl	APIC_ICR_LOW(%esi),%eax
	andl	$0xfff00000,%eax
	orl	$0x0000069f,%eax	/* do a STARTUP IPI */
	movl	%eax,APIC_ICR_LOW(%esi)

3:	movl	APIC_ICR_LOW(%esi),%eax	/* wait for pending status end */
	andl 	$0x00001000,%eax
	jnz 	3b

	xorl	%ebx, %ebx
	movl	$0x1000, %eax /* Delay a bit */
3:	decl	%eax
	cmpl	%eax, %ebx
	jnz	3b

4:	movl	APIC_ICR_LOW(%esi),%eax	/* wait for pending status end */
	andl 	$0x00001000,%eax
	jnz 	4b

	movl	APIC_ICR_HI(%esi),%eax
	andl	$0xf0ffffff,%eax /* mask out ID bits */
	orl	$CPUNBR,%eax
	movl	%eax,APIC_ICR_HI(%esi)
	movl	APIC_ICR_LOW(%esi),%eax
	andl	$0xfff00000,%eax
	orl	$0x0000069f,%eax	/* do ANOTHER STARTUP IPI */
	movl	%eax,APIC_ICR_LOW(%esi)

3:	movl	APIC_ICR_LOW(%esi),%eax	/* wait for pending status end */
	andl 	$0x00001000,%eax
	jnz 	3b

	popl	%ebx
	popl	%esi

	ret
