/*
 * Local APIC definitions.
 *
 * $Id: apic.h,v 1.1 1996/05/06 20:13:28 peter Exp $
 */

/*
 * Pentium P54C+ Build-in APIC
 * (Advanced programmable Interrupt Controller)
 * 
 * Base Address of Build-in APIC in memory location
 * is 0xfee00000.
 * 
 * Map of APIC REgisters:
 * 
 * Offset (hex)    Description                     Read/Write state
 * 000             Reserved
 * 010             Reserved
 * 020 ID          Local APIC ID                   R/W
 * 030 VER         Local APIC Version              R
 * 040             Reserved
 * 050             Reserved
 * 060             Reserved
 * 070             Reserved
 * 080             Task Priority Register          R/W
 * 090             Arbitration Priority Register   R
 * 0A0             Processor Priority Register     R
 * 0B0             EOI Register                    W
 * 0C0 RRR         Remote read                     R
 * 0D0             Logical Destination             R/W
 * 0E0             Destination Format Register     0..27 R;  28..31 R/W
 * 0F0 SVR         Spurious Interrupt Vector Reg.  0..3  R;  4..9   R/W
 * 100             ISR  000-031                    R
 * 110             ISR  032-063                    R
 * 120             ISR  064-095                    R
 * 130             ISR  095-128                    R
 * 140             ISR  128-159                    R
 * 150             ISR  160-191                    R
 * 160             ISR  192-223                    R
 * 170             ISR  224-255                    R
 * 180             TMR  000-031                    R
 * 190             TMR  032-063                    R
 * 1A0             TMR  064-095                    R
 * 1B0             TMR  095-128                    R
 * 1C0             TMR  128-159                    R
 * 1D0             TMR  160-191                    R
 * 1E0             TMR  192-223                    R
 * 1F0             TMR  224-255                    R
 * 200             IRR  000-031                    R
 * 210             IRR  032-063                    R
 * 220             IRR  064-095                    R
 * 230             IRR  095-128                    R
 * 240             IRR  128-159                    R
 * 250             IRR  160-191                    R
 * 260             IRR  192-223                    R
 * 270             IRR  224-255                    R
 * 280             Error Status Register           R
 * 290             Reserved
 * 2A0             Reserved
 * 2B0             Reserved
 * 2C0             Reserved
 * 2D0             Reserved
 * 2E0             Reserved
 * 2F0             Reserved
 * 300 ICR_LOW     Interrupt Command Reg. (0-31)   R/W
 * 310 ICR_HI      Interrupt Command Reg. (32-63)  R/W
 * 320             Local Vector Table (Timer)      R/W
 * 330             Reserved
 * 340             Reserved
 * 350 LVT1        Local Vector Table (LINT0)      R/W
 * 360 LVT2        Local Vector Table (LINT1)      R/W
 * 370 LVT3        Local Vector Table (ERROR)      R/W
 * 380             Initial Count Reg. for Timer    R/W
 * 390             Current Count of Timer          R
 * 3A0             Reserved
 * 3B0             Reserved
 * 3C0             Reserved
 * 3D0             Reserved
 * 3E0             Timer Divide Configuration Reg. R/W
 * 3F0             Reserved
 */

#ifndef _MACHINE_APIC_H_
#define _MACHINE_APIC_H_

/* Physical locations of APIC's */
#define APIC_BASE	0xfee00000
#define IO_APIC_BASE	0xfec00000

/* APIC ID mask */
#define APIC_ID_MASK	0x0f000000

#ifdef LOCORE

/* Offsets from _apic_base */
#define APIC_ID		0x020
#define APIC_VER	0x030
#define APIC_RRR	0x0c0
#define APIC_SVR	0x0f0
#define APIC_ICR_LOW	0x300
#define APIC_ICR_HI	0x310
#define APIC_LVT1	0x350
#define APIC_LVT2	0x360
#define APIC_LVT3	0x370


#else /* !LOCORE */

volatile unsigned int *apic_base;

/* Offsets in apic_base[] */
#define APIC_ID		(0x020/4)
#define APIC_VER	(0x030/4)
#define APIC_RRR	(0x0c0/4)
#define APIC_SVR	(0x0f0/4)
#define APIC_ICR_LOW	(0x300/4)
#define APIC_ICR_HI	(0x310/4)
#define APIC_LVT1	(0x350/4)
#define APIC_LVT2	(0x360/4)
#define APIC_LVT3	(0x370/4)

#endif /* LOCORE */
#endif /* _MACHINE_APIC_H */
